/* Created on 25-ago-2006 */
package ita.trem.app;

import ita.trem.gui.ProgressBarDialog;
import ita.trem.gui.TremDownloaderGui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;

import java.security.MessageDigest;

import java.util.LinkedList;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;


public class TremDownloader extends Thread {

	public static final String listUrl = "http://www.capponcino.it/tremulous/dl/pk3/";
	private String path;
	private boolean toRun;
	private TremDownloaderGui gui;
        private List<Mappa> listaMappeInput;
	
	public TremDownloader(String pathToUpdate,TremDownloaderGui gui, List<Mappa> listaMappeInput) {
		this.path = pathToUpdate;
		this.toRun = true;
		this.gui = gui;
                this.listaMappeInput = listaMappeInput;
	}

	public synchronized void setToRun(boolean b) {
		this.toRun = b;
	}
	
	public synchronized boolean isToRun() {
		return this.toRun;
	}
	
	public void run() {
		setToRun(true);
		gui.enableStop();
		gui.setLog("========================" + (char) 10 + (char) 13);
		URL url = null;
		URLConnection urlconn = null;
		
		/* Connessione al server */
		if (isToRun()) {
			gui.setLog("Connecting to server..." + (char) 10 + (char) 13);
			
			try {
				url = new URL(listUrl + "maps.log");
			} catch (MalformedURLException e) { /* Impossibile */ }
			
			try {
				urlconn = url.openConnection();
			} catch (IOException e) {
				gui.setLog("Connection to server failed. Aborting..." + (char) 10 + (char) 13);
				gui.disableStop();
				return;
			}
			gui.setServerConnection(true);
		} else {
			gui.setLog("Abort by user request." + (char) 10 + (char) 13);
			gui.disableStop();
			return;
		}
                
                List<Mappa> listaMappe = null;
                //Se la lista non arriva gi passata in ingresso la creo prendendo tutte quelle sul server ( come era nel sorgente originale)
                if(listaMappeInput == null)  {
                
                  /* Apertura stream verso lista mappe */
                  InputStream is = null;
                  if (isToRun()) {
                          gui.setLog("Opening stream..." + (char) 10 + (char) 13);
                          try {
                                  is = urlconn.getInputStream();
                          } catch (IOException e) {
                                  gui.setLog("List retrieval failed. Aborting..." + (char) 10 + (char) 13);
                                  gui.disableStop();
                                  return;
                          }
                          gui.setListFetch(true);
                  } else {
                          gui.setLog("Abort by user request." + (char) 10 + (char) 13);
                          gui.disableStop();
                          return;
                  }
                  
                  BufferedReader buff = new BufferedReader(new InputStreamReader(is));
                
                  listaMappe = new LinkedList<Mappa>();
                  /* Creazione lista mappe */
                  if (isToRun()) {
                          gui.setLog("Processing list..." + (char) 10 + (char) 13);
                          String line = null;
                          try {
                                  do {
                                          String nome;
                                          String vers;
                                          String fileServer;
                                          String pkfile;
                                          String sum;
                                          nome = buff.readLine();
                                          vers = buff.readLine();
                                          fileServer = buff.readLine();
                                          pkfile = buff.readLine();
                                          sum=buff.readLine();
                                          /* Crea oggetto mappa */
                                          listaMappe.add(new Mappa(nome,vers,pkfile,fileServer,sum));
                                          /* */
                                          line = buff.readLine();
                                  } while (line != null);
                          } catch (IOException e) {
                                  if (listaMappe.size()==0) {
                                          gui.setLog("List processing failed. Aborting..." + (char) 10 + (char) 13);
                                          gui.disableStop();
                                          return;
                                  } else {
                                          gui.setLog("Some troubles on parsing file, but some maps found..." + (char) 10 + (char) 13);
                                  }
                          }
                          gui.setListProcess(true);
                  } else {
                          gui.setLog("Abort by user request." + (char) 10 + (char) 13);
                          gui.disableStop();
                          return;
                  }
                
                }
                else{
                  listaMappe = listaMappeInput;
                  gui.setListFetch(true);
                  gui.setListProcess(true);
                }
                
		/* Aggiornamento mappe */
		if (isToRun()) {
			gui.setLog("Updating maps..." + (char) 10 + (char) 13);
			updateMaps(listaMappe);
			gui.setMapUpdate(true);
		} else {
			gui.setLog("Abort by user request." + (char) 10 + (char) 13);
			gui.disableStop();
			return;
		}
		
		if (isToRun()) {
			gui.setDeleteTemp(true);
		} else {
			gui.setLog("Abort by user request." + (char) 10 + (char) 13);
			gui.disableStop();
			return;
		}
		
		gui.setLog("All operations performed." + (char) 10 + (char) 13);
		gui.setDone(true);
		gui.disableStop();
	}

	private void updateMaps(List<Mappa> listaMappe) {
		for(int i = 0;i<listaMappe.size() && isToRun();i++) {
			Mappa m = listaMappe.get(i);
			/* Controlla se il file pk3 esiste nella dir specificata */
			File locale = new File(this.path + "/" + m.getNomeFile());
			
			/* Se si, avvisa che il file e' aggiornato */
			if (locale.exists()) {
				/* Controlla la corrispondenza MD5 */
				String md5sum = calcMD5(locale);
				if (md5sum.equals(m.getMD5()))
					this.gui.setLog(m.getNome() + " is already up-to-date" + (char) 10 + (char) 13);
				else {
					this.gui.setLog(m.getNome() + " is to update (MD5 failure)"  + (char) 10 + (char) 13);
					this.gui.setLog("Local md5: " + md5sum + " - Remote md5: " + m.getMD5()  + (char) 10 + (char) 13);
					locale.delete();
					update(m);
				}
			} else {
				this.gui.setLog(m.getNome() + " is to update"  + (char) 10 + (char) 13);
				update(m);
			}
		}
	}

	private void update(Mappa m) {
		/* Se no, controlla l'estensione del file sul server */
		int dot = m.getFileSulServer().lastIndexOf('.');
		String ext = m.getFileSulServer().substring(dot);
		/* Se e' pk3, scaricalo semplicemente nella dir specificata */
		try {
			if (ext.equals(".pk3")) {
				download(m);
			} else {
			/* Altrimenti scaricalo e decomprimilo */
				downloadZip(m);
			}
		} catch (Exception e) {
			this.gui.setLog(m.getNome() + " not downloaded. Skipping to next map..." + (char) 10 + (char) 13);
		}
	}

	private String calcMD5(File locale) {
		try {
			MessageDigest md5 = MessageDigest.getInstance("MD5");
			byte[] b = new byte[4*1024];
			InputStream insock = new FileInputStream(locale);
	
			int lunghezza = insock.read(b);
		
			while (lunghezza > 0 && isToRun()) {
					md5.update(b,0,lunghezza);
					lunghezza = insock.read(b);
			}
			insock.close();
			byte[] res = md5.digest();
			String output = formatta(res);
			return output;
		} catch (Exception e) { e.printStackTrace(); return ""; }
	}

	public static String formatta(byte[] res) {
		StringBuffer sb = new StringBuffer();
		for (int i = 0;i<res.length;i++) {
			String part = Integer.toHexString(new Integer(res[i]));
			if (part.length()==1)
				sb.append("0" + part);
			else
				sb.append(part.substring(part.length()-2));
		}
		return sb.toString();
	}

	private void downloadZip(Mappa m) {
		/* decomprimi il file */
		 try {
			/* scarica il file zip */
                        download(m);
		        // Open the ZIP file
		        String inFilename = this.path + "/" + m.getFileSulServer();
		        ZipInputStream in = new ZipInputStream(new FileInputStream(inFilename));
		  
		        // Search for pk3 file
		        boolean found = false;
		        
		        while(!found) {
		        	try {
		        	// Get the entry
			        ZipEntry entry = in.getNextEntry();
			        int dot = entry.getName().lastIndexOf('.');
					String ext = entry.getName().substring(dot);
			        if(ext.equals(".pk3")) {
			        	found = true;
			        } else {
			        	this.gui.setLog(entry.getName() + " not a pk3 file. Try next.." + (char) 10 + (char)13);
			        }
		        	} catch (IOException e) {
		        		this.gui.setLog("Error while processing file " + m.getFileSulServer() + ". Skipping..." + (char) 10 + (char)13);
		        		found = true;
		        		throw new Exception("No pk3 in zip!");
		        	}
		        }
		        // Open the output file
		        String outFilename = this.path + "/" + m.getNomeFile();
		        OutputStream out = new FileOutputStream(outFilename);
		    
		        // Transfer bytes from the ZIP file to the output file
		        byte[] buf = new byte[2048];
		        int len;
		        while ((len = in.read(buf)) > 0) {
		            out.write(buf, 0, len);
		        }
		        // Close the streams
		        out.close();
		        in.close();
		    } catch (Exception e) {
		    	this.gui.setLog("There was a problem decompressing " + m.getNome() + ". Skipping..."  + (char) 10 + (char) 13);
		    } finally {
		    	// Delete zip
		    	File f = new File(this.path + "/" + m.getFileSulServer());
		    	f.delete();
		    }
		    
	}

	private void download(Mappa m) throws Exception {
		try {
			this.gui.setLog("Download in progress..."+ (char) 10 + (char) 13);
			URL mappaUrl = new URL(listUrl + m.getFileSulServer());
			File locale = new File(this.path + "/" + m.getFileSulServer());
			locale.createNewFile();
			
			byte[] b = new byte[4*1024];
			FileOutputStream fileWrite = new FileOutputStream(locale);
			InputStream insock = mappaUrl.openStream();
                        
                        //Il valore massimo per la JProgressBar  la lunghezza del file in bytes
                        URLConnection conn = mappaUrl.openConnection();
                        int max =conn.getContentLength();
                        int totale = 0;
                        
                        
                        ProgressBarDialog pbdlg =  ProgressBarDialog.getProgressBarDialog (gui.getJFrame(), m.getNome(), max, ProgressBarDialog.DOWNLOAD_PROGRESS_DIALOG);
                        pbdlg.addCancelButtonActionListener(new ActionListener(){
                          public void actionPerformed(ActionEvent e){
                            setToRun(false);
                          }
                        });
	
			int lunghezza = insock.read(b);
                        totale += lunghezza;
			int passate = 0;
			while (lunghezza > 0 && isToRun()) {
					fileWrite.write(b,0,lunghezza);
					lunghezza = insock.read(b);

                                        // Per ogni ciclo di lettura aggiorno la progressbar 
                                        totale += lunghezza;
                                        pbdlg.setBarValue(totale);
                                        
					passate++;
					if (passate==140) {
						this.gui.setLog("#");
						passate = 0;
					}
			}
			this.gui.setLog("#" + (char) 10 + (char) 13);
			
                        pbdlg.setVisible(false);
                        
			insock.close();
			fileWrite.close();
			
			if (isToRun()==false) {
				this.gui.setLog("Aborting file transfer..."+ (char) 10 + (char) 13);
				locale.delete();
				throw new IOException();
			}
			
		} catch (Exception e) {
			this.gui.setLog("There was a problem downloading " + m.getNome() + ". Skipping..."  + (char) 10 + (char) 13);
			throw e;
		}
	}

}
